# Copyright 2020 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Channelz debugging tool for xDS test client/server.

This is intended as a debugging / local development helper and not executed
as a part of interop test suites.

Typical usage examples:

    # Show channel and server socket pair
    python -m bin.run_channelz --flagfile=config/local-dev.cfg

    # Evaluate setup for different security configurations
    python -m bin.run_channelz --flagfile=config/local-dev.cfg --security=tls
    python -m bin.run_channelz --flagfile=config/local-dev.cfg --security=mtls_error

    # More information and usage options
    python -m bin.run_channelz --helpfull
"""
import hashlib

from absl import app
from absl import flags
from absl import logging

from bin.lib import common
from framework import xds_flags
from framework import xds_k8s_flags
from framework.infrastructure import gcp
from framework.infrastructure import k8s
from framework.rpc import grpc_channelz
from framework.test_app import client_app
from framework.test_app import server_app

# Flags
_SECURITY = flags.DEFINE_enum(
    "security",
    default=None,
    enum_values=[
        "mtls",
        "tls",
        "plaintext",
        "mtls_error",
        "server_authz_error",
    ],
    help="Show info for a security setup",
)
flags.adopt_module_key_flags(xds_flags)
flags.adopt_module_key_flags(xds_k8s_flags)
# Running outside of a test suite, so require explicit resource_suffix.
flags.mark_flag_as_required(xds_flags.RESOURCE_SUFFIX.name)
flags.register_validator(
    xds_flags.SERVER_XDS_PORT.name,
    lambda val: val > 0,
    message=(
        "Run outside of a test suite, must provide"
        " the exact port value (must be greater than 0)."
    ),
)

logger = logging.get_absl_logger()

# Type aliases
_Channel = grpc_channelz.Channel
_Socket = grpc_channelz.Socket
_ChannelState = grpc_channelz.ChannelState
_XdsTestServer = server_app.XdsTestServer
_XdsTestClient = client_app.XdsTestClient


def debug_cert(cert):
    if not cert:
        return "<missing>"
    sha1 = hashlib.sha1(cert)
    return f"sha1={sha1.hexdigest()}, len={len(cert)}"


def debug_sock_tls(tls):
    return (
        f"local:  {debug_cert(tls.local_certificate)}\n"
        f"remote: {debug_cert(tls.remote_certificate)}"
    )


def get_deployment_pods(k8s_ns, deployment_name):
    deployment = k8s_ns.get_deployment(deployment_name)
    return k8s_ns.list_deployment_pods(deployment)


def debug_security_setup_negative(test_client):
    """Debug negative cases: mTLS Error, Server AuthZ error

    1) mTLS Error: Server expects client mTLS cert,
       but client configured only for TLS.
    2) AuthZ error: Client does not authorize server because of mismatched
       SAN name.
    """
    # Client side.
    client_correct_setup = True
    channel: _Channel = test_client.wait_for_server_channel_state(
        state=_ChannelState.TRANSIENT_FAILURE
    )
    try:
        subchannel, *subchannels = list(
            test_client.channelz.list_channel_subchannels(channel)
        )
    except ValueError:
        print(
            "Client setup fail: subchannel not found. "
            "Common causes: test client didn't connect to TD; "
            "test client exhausted retries, and closed all subchannels."
        )
        return

    # Client must have exactly one subchannel.
    logger.debug("Found subchannel, %s", subchannel)
    if subchannels:
        client_correct_setup = False
        print(f"Unexpected subchannels {subchannels}")
    subchannel_state: _ChannelState = subchannel.data.state.state
    if subchannel_state is not _ChannelState.TRANSIENT_FAILURE:
        client_correct_setup = False
        print(
            "Subchannel expected to be in "
            "TRANSIENT_FAILURE, same as its channel"
        )

    # Client subchannel must have no sockets.
    sockets = list(test_client.channelz.list_subchannels_sockets(subchannel))
    if sockets:
        client_correct_setup = False
        print(f"Unexpected subchannel sockets {sockets}")

    # Results.
    if client_correct_setup:
        print(
            "Client setup pass: the channel "
            "to the server has exactly one subchannel "
            "in TRANSIENT_FAILURE, and no sockets"
        )


def debug_security_setup_positive(test_client, test_server):
    """Debug positive cases: mTLS, TLS, Plaintext."""
    test_client.wait_for_active_server_channel()
    client_sock: _Socket = test_client.get_active_server_channel_socket()
    server_sock: _Socket = test_server.get_server_socket_matching_client(
        client_sock
    )

    server_tls = server_sock.security.tls
    client_tls = client_sock.security.tls

    print(f"\nServer certs:\n{debug_sock_tls(server_tls)}")
    print(f"\nClient certs:\n{debug_sock_tls(client_tls)}")
    print()

    if server_tls.local_certificate:
        eq = server_tls.local_certificate == client_tls.remote_certificate
        print(f"(TLS)  Server local matches client remote: {eq}")
    else:
        print("(TLS)  Not detected")

    if server_tls.remote_certificate:
        eq = server_tls.remote_certificate == client_tls.local_certificate
        print(f"(mTLS) Server remote matches client local: {eq}")
    else:
        print("(mTLS) Not detected")


def debug_basic_setup(test_client, test_server):
    """Show channel and server socket pair"""
    test_client.wait_for_active_server_channel()
    client_sock: _Socket = test_client.get_active_server_channel_socket()
    server_sock: _Socket = test_server.get_server_socket_matching_client(
        client_sock
    )

    logger.debug("Client socket: %s\n", client_sock)
    logger.debug("Matching server socket: %s\n", server_sock)


def main(argv):
    if len(argv) > 1:
        raise app.UsageError("Too many command-line arguments.")

    # Must be called before KubernetesApiManager or GcpApiManager init.
    xds_flags.set_socket_default_timeout_from_flag()

    # Flags.
    should_port_forward: bool = xds_k8s_flags.DEBUG_USE_PORT_FORWARDING.value
    is_secure: bool = bool(_SECURITY.value)

    # Setup.
    gcp_api_manager = gcp.api.GcpApiManager()
    k8s_api_manager = k8s.KubernetesApiManager(xds_k8s_flags.KUBE_CONTEXT.value)

    # Server.
    server_namespace = common.make_server_namespace(k8s_api_manager)
    server_runner = common.make_server_runner(
        server_namespace,
        gcp_api_manager,
        port_forwarding=should_port_forward,
        mode="secure",
    )
    # Find server pod.
    server_pod: k8s.V1Pod = common.get_server_pod(
        server_runner, xds_flags.SERVER_NAME.value
    )

    # Client
    client_namespace = common.make_client_namespace(k8s_api_manager)
    client_runner = common.make_client_runner(
        client_namespace,
        gcp_api_manager,
        port_forwarding=should_port_forward,
        mode="secure",
    )
    # Find client pod.
    client_pod: k8s.V1Pod = common.get_client_pod(
        client_runner, xds_flags.CLIENT_NAME.value
    )

    # Ensure port forwarding stopped.
    common.register_graceful_exit(server_runner, client_runner)

    # Create server app for the server pod.
    test_server: _XdsTestServer = common.get_test_server_for_pod(
        server_runner,
        server_pod,
        test_port=xds_flags.SERVER_PORT.value,
        secure_mode=is_secure,
    )
    test_server.set_xds_address(
        xds_flags.SERVER_XDS_HOST.value, xds_flags.SERVER_XDS_PORT.value
    )

    # Create client app for the client pod.
    test_client: _XdsTestClient = common.get_test_client_for_pod(
        client_runner, client_pod, server_target=test_server.xds_uri
    )

    with test_client, test_server:
        if _SECURITY.value in ("mtls", "tls", "plaintext"):
            debug_security_setup_positive(test_client, test_server)
        elif _SECURITY.value in ("mtls_error", "server_authz_error"):
            debug_security_setup_negative(test_client)
        else:
            debug_basic_setup(test_client, test_server)

    logger.info("SUCCESS!")


if __name__ == "__main__":
    app.run(main)



# Copyright 2022 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Clean up GKE namespaces leaked by the tests."""

from absl import app

from bin.cleanup import cleanup
from framework import xds_flags
from framework import xds_k8s_flags


def main(argv):
    if len(argv) > 1:
        raise app.UsageError("Too many command-line arguments.")
    cleanup.load_keep_config()

    # Must be called before KubernetesApiManager or GcpApiManager init.
    xds_flags.set_socket_default_timeout_from_flag()

    project: str = xds_flags.PROJECT.value
    network: str = xds_flags.NETWORK.value
    gcp_service_account: str = xds_k8s_flags.GCP_SERVICE_ACCOUNT.value
    dry_run: bool = cleanup.DRY_RUN.value

    cleanup.find_and_remove_leaked_k8s_resources(
        dry_run, project, network, gcp_service_account
    )


if __name__ == "__main__":
    app.run(main)



#!/usr/bin/env python3

# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import datetime
import os
import re
import subprocess
import sys

# find our home
ROOT = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), "../.."))
os.chdir(ROOT)

# parse command line
argp = argparse.ArgumentParser(description="copyright checker")
argp.add_argument(
    "-o", "--output", default="details", choices=["list", "details"]
)
argp.add_argument("-s", "--skips", default=0, action="store_const", const=1)
argp.add_argument("-a", "--ancient", default=0, action="store_const", const=1)
argp.add_argument("--precommit", action="store_true")
argp.add_argument("--fix", action="store_true")
args = argp.parse_args()

# open the license text
with open("NOTICE.txt") as f:
    LICENSE_NOTICE = f.read().splitlines()

# license format by file extension
# key is the file extension, value is a format string
# that given a line of license text, returns what should
# be in the file
LICENSE_PREFIX_RE = {
    ".bat": r"@rem\s*",
    ".c": r"\s*(?://|\*)\s*",
    ".cc": r"\s*(?://|\*)\s*",
    ".h": r"\s*(?://|\*)\s*",
    ".m": r"\s*\*\s*",
    ".mm": r"\s*\*\s*",
    ".php": r"\s*\*\s*",
    ".js": r"\s*\*\s*",
    ".py": r"#\s*",
    ".pyx": r"#\s*",
    ".pxd": r"#\s*",
    ".pxi": r"#\s*",
    ".rb": r"#\s*",
    ".sh": r"#\s*",
    ".proto": r"//\s*",
    ".cs": r"//\s*",
    ".mak": r"#\s*",
    ".bazel": r"#\s*",
    ".bzl": r"#\s*",
    "Makefile": r"#\s*",
    "Dockerfile": r"#\s*",
    "BUILD": r"#\s*",
}

# The key is the file extension, while the value is a tuple of fields
# (header, prefix, footer).
# For example, for javascript multi-line comments, the header will be '/*', the
# prefix will be '*' and the footer will be '*/'.
# If header and footer are irrelevant for a specific file extension, they are
# set to None.
LICENSE_PREFIX_TEXT = {
    ".bat": (None, "@rem", None),
    ".c": (None, "//", None),
    ".cc": (None, "//", None),
    ".h": (None, "//", None),
    ".m": ("/**", " *", " */"),
    ".mm": ("/**", " *", " */"),
    ".php": ("/**", " *", " */"),
    ".js": ("/**", " *", " */"),
    ".py": (None, "#", None),
    ".pyx": (None, "#", None),
    ".pxd": (None, "#", None),
    ".pxi": (None, "#", None),
    ".rb": (None, "#", None),
    ".sh": (None, "#", None),
    ".proto": (None, "//", None),
    ".cs": (None, "//", None),
    ".mak": (None, "#", None),
    ".bazel": (None, "#", None),
    ".bzl": (None, "#", None),
    "Makefile": (None, "#", None),
    "Dockerfile": (None, "#", None),
    "BUILD": (None, "#", None),
}

_EXEMPT = frozenset(
    (
        # Generated protocol compiler output.
        "examples/python/helloworld/helloworld_pb2.py",
        "examples/python/helloworld/helloworld_pb2_grpc.py",
        "examples/python/multiplex/helloworld_pb2.py",
        "examples/python/multiplex/helloworld_pb2_grpc.py",
        "examples/python/multiplex/route_guide_pb2.py",
        "examples/python/multiplex/route_guide_pb2_grpc.py",
        "examples/python/route_guide/route_guide_pb2.py",
        "examples/python/route_guide/route_guide_pb2_grpc.py",
        # Generated doxygen config file
        "tools/doxygen/Doxyfile.php",
        # An older file originally from outside gRPC.
        "src/php/tests/bootstrap.php",
        # census.proto copied from github
        "tools/grpcz/census.proto",
        # status.proto copied from googleapis
        "src/proto/grpc/status/status.proto",
        # Gradle wrappers used to build for Android
        "examples/android/helloworld/gradlew.bat",
        "src/android/test/interop/gradlew.bat",
        # Designer-generated source
        "examples/csharp/HelloworldXamarin/Droid/Resources/Resource.designer.cs",
        "examples/csharp/HelloworldXamarin/iOS/ViewController.designer.cs",
        # BoringSSL generated header. It has commit version information at the head
        # of the file so we cannot check the license info.
        "src/boringssl/boringssl_prefix_symbols.h",
    )
)

_ENFORCE_CPP_STYLE_COMMENT_PATH_PREFIX = tuple(
    [
        "include/grpc++/",
        "include/grpcpp/",
        "src/core/",
        "src/cpp/",
        "test/core/",
        "test/cpp/",
        "fuzztest/",
    ]
)

RE_YEAR = (
    r"Copyright (?P<first_year>[0-9]+\-)?(?P<last_year>[0-9]+) ([Tt]he )?gRPC"
    r" [Aa]uthors(\.|)"
)
RE_LICENSE = dict(
    (
        k,
        r"\n".join(
            LICENSE_PREFIX_RE[k]
            + (RE_YEAR if re.search(RE_YEAR, line) else re.escape(line))
            for line in LICENSE_NOTICE
        ),
    )
    for k, v in list(LICENSE_PREFIX_RE.items())
)

RE_C_STYLE_COMMENT_START = r"^/\*\s*\n"
RE_C_STYLE_COMMENT_OPTIONAL_LINE = r"(?:\s*\*\s*\n)*"
RE_C_STYLE_COMMENT_END = r"\s*\*/"
RE_C_STYLE_COMMENT_LICENSE = (
    RE_C_STYLE_COMMENT_START
    + RE_C_STYLE_COMMENT_OPTIONAL_LINE
    + r"\n".join(
        r"\s*(?:\*)\s*"
        + (RE_YEAR if re.search(RE_YEAR, line) else re.escape(line))
        for line in LICENSE_NOTICE
    )
    + r"\n"
    + RE_C_STYLE_COMMENT_OPTIONAL_LINE
    + RE_C_STYLE_COMMENT_END
)
RE_CPP_STYLE_COMMENT_LICENSE = r"\n".join(
    r"\s*(?://)\s*" + (RE_YEAR if re.search(RE_YEAR, line) else re.escape(line))
    for line in LICENSE_NOTICE
)

YEAR = datetime.datetime.now().year

LICENSE_YEAR = f"Copyright {YEAR} gRPC authors."


def join_license_text(header, prefix, footer, notice):
    text = (header + "\n") if header else ""

    def add_prefix(prefix, line):
        # Don't put whitespace between prefix and empty line to avoid having
        # trailing whitespaces.
        return prefix + ("" if len(line) == 0 else " ") + line

    text += "\n".join(
        add_prefix(prefix, (LICENSE_YEAR if re.search(RE_YEAR, line) else line))
        for line in LICENSE_NOTICE
    )
    text += "\n"
    if footer:
        text += footer + "\n"
    return text


LICENSE_TEXT = dict(
    (
        k,
        join_license_text(
            LICENSE_PREFIX_TEXT[k][0],
            LICENSE_PREFIX_TEXT[k][1],
            LICENSE_PREFIX_TEXT[k][2],
            LICENSE_NOTICE,
        ),
    )
    for k, v in list(LICENSE_PREFIX_TEXT.items())
)

if args.precommit:
    FILE_LIST_COMMAND = (
        "git status -z | grep -Poz '(?<=^[MARC][MARCD ] )[^\s]+'"
    )
else:
    FILE_LIST_COMMAND = (
        "git ls-tree -r --name-only -r HEAD | "
        "grep -v ^third_party/ |"
        'grep -v "\(ares_config.h\|ares_build.h\)"'
    )


def load(name):
    with open(name) as f:
        return f.read()


def save(name, text):
    with open(name, "w") as f:
        f.write(text)


assert re.search(RE_LICENSE["Makefile"], load("Makefile"))


def log(cond, why, filename):
    if not cond:
        return
    if args.output == "details":
        print(("%s: %s" % (why, filename)))
    else:
        print(filename)


def write_copyright(license_text, file_text, filename):
    shebang = ""
    lines = file_text.split("\n")
    if lines and lines[0].startswith("#!"):
        shebang = lines[0] + "\n"
        file_text = file_text[len(shebang) :]

    rewritten_text = shebang + license_text + "\n" + file_text
    with open(filename, "w") as f:
        f.write(rewritten_text)


def replace_copyright(license_text, file_text, filename):
    m = re.search(RE_C_STYLE_COMMENT_LICENSE, text)
    if m:
        rewritten_text = license_text + file_text[m.end() :]
        with open(filename, "w") as f:
            f.write(rewritten_text)
        return True
    return False


# scan files, validate the text
ok = True
filename_list = []
try:
    filename_list = (
        subprocess.check_output(FILE_LIST_COMMAND, shell=True)
        .decode()
        .splitlines()
    )
except subprocess.CalledProcessError:
    sys.exit(0)

for filename in filename_list:
    enforce_cpp_style_comment = False
    if filename in _EXEMPT:
        continue
    # Skip check for upb generated code.
    if (
        filename.endswith(".upb.h")
        or filename.endswith(".upb.c")
        or filename.endswith(".upbdefs.h")
        or filename.endswith(".upbdefs.c")
    ):
        continue
    ext = os.path.splitext(filename)[1]
    base = os.path.basename(filename)
    if filename.startswith(_ENFORCE_CPP_STYLE_COMMENT_PATH_PREFIX) and ext in [
        ".cc",
        ".h",
    ]:
        enforce_cpp_style_comment = True
        re_license = RE_CPP_STYLE_COMMENT_LICENSE
        license_text = LICENSE_TEXT[ext]
    elif ext in RE_LICENSE:
        re_license = RE_LICENSE[ext]
        license_text = LICENSE_TEXT[ext]
    elif base in RE_LICENSE:
        re_license = RE_LICENSE[base]
        license_text = LICENSE_TEXT[base]
    else:
        log(args.skips, "skip", filename)
        continue
    try:
        text = load(filename)
    except:
        continue
    m = re.search(re_license, text)
    if m:
        pass
    elif enforce_cpp_style_comment:
        log(
            1,
            "copyright missing or does not use cpp-style copyright header",
            filename,
        )
        if args.fix:
            # Attempt fix: search for c-style copyright header and replace it
            # with cpp-style copyright header. If that doesn't work
            # (e.g. missing copyright header), write cpp-style copyright header.
            if not replace_copyright(license_text, text, filename):
                write_copyright(license_text, text, filename)
        ok = False
    elif "DO NOT EDIT" not in text:
        if args.fix:
            write_copyright(license_text, text, filename)
            log(1, "copyright missing (fixed)", filename)
        else:
            log(1, "copyright missing", filename)
        ok = False

if not ok and not args.fix:
    print(
        "You may use following command to automatically fix copyright headers:"
    )
    print("    tools/distrib/check_copyright.py --fix")

sys.exit(0 if ok else 1)



# Copyright 2018 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests experimental TLS Session Resumption API"""

import logging
import pickle
import unittest

import grpc
from grpc import _channel
from grpc.experimental import session_cache

from tests.unit import resources
from tests.unit import test_common

_REQUEST = b"\x00\x00\x00"
_RESPONSE = b"\x00\x00\x00"

_UNARY_UNARY = "/test/UnaryUnary"

_SERVER_HOST_OVERRIDE = "foo.test.google.fr"
_ID = "id"
_ID_KEY = "id_key"
_AUTH_CTX = "auth_ctx"

_PRIVATE_KEY = resources.private_key()
_CERTIFICATE_CHAIN = resources.certificate_chain()
_TEST_ROOT_CERTIFICATES = resources.test_root_certificates()
_SERVER_CERTS = ((_PRIVATE_KEY, _CERTIFICATE_CHAIN),)
_PROPERTY_OPTIONS = (
    (
        "grpc.ssl_target_name_override",
        _SERVER_HOST_OVERRIDE,
    ),
)


def handle_unary_unary(request, servicer_context):
    return pickle.dumps(
        {
            _ID: servicer_context.peer_identities(),
            _ID_KEY: servicer_context.peer_identity_key(),
            _AUTH_CTX: servicer_context.auth_context(),
        }
    )


def start_secure_server():
    handler = grpc.method_handlers_generic_handler(
        "test",
        {"UnaryUnary": grpc.unary_unary_rpc_method_handler(handle_unary_unary)},
    )
    server = test_common.test_server()
    server.add_generic_rpc_handlers((handler,))
    server_cred = grpc.ssl_server_credentials(_SERVER_CERTS)
    port = server.add_secure_port("[::]:0", server_cred)
    server.start()

    return server, port


class SSLSessionCacheTest(unittest.TestCase):
    def _do_one_shot_client_rpc(
        self, channel_creds, channel_options, port, expect_ssl_session_reused
    ):
        channel = grpc.secure_channel(
            "localhost:{}".format(port), channel_creds, options=channel_options
        )
        response = channel.unary_unary(_UNARY_UNARY)(_REQUEST)
        auth_data = pickle.loads(response)
        self.assertEqual(
            expect_ssl_session_reused,
            auth_data[_AUTH_CTX]["ssl_session_reused"],
        )
        channel.close()

    def testSSLSessionCacheLRU(self):
        server_1, port_1 = start_secure_server()

        cache = session_cache.ssl_session_cache_lru(1)
        channel_creds = grpc.ssl_channel_credentials(
            root_certificates=_TEST_ROOT_CERTIFICATES
        )
        channel_options = _PROPERTY_OPTIONS + (
            ("grpc.ssl_session_cache", cache),
        )

        # Initial connection has no session to resume
        self._do_one_shot_client_rpc(
            channel_creds,
            channel_options,
            port_1,
            expect_ssl_session_reused=[b"false"],
        )

        # Connection to server_1 resumes from initial session
        self._do_one_shot_client_rpc(
            channel_creds,
            channel_options,
            port_1,
            expect_ssl_session_reused=[b"true"],
        )

        # Connection to a different server with the same name overwrites the cache entry
        server_2, port_2 = start_secure_server()
        self._do_one_shot_client_rpc(
            channel_creds,
            channel_options,
            port_2,
            expect_ssl_session_reused=[b"false"],
        )
        self._do_one_shot_client_rpc(
            channel_creds,
            channel_options,
            port_2,
            expect_ssl_session_reused=[b"true"],
        )
        server_2.stop(None)

        # Connection to server_1 now falls back to full TLS handshake
        self._do_one_shot_client_rpc(
            channel_creds,
            channel_options,
            port_1,
            expect_ssl_session_reused=[b"false"],
        )

        # Re-creating server_1 causes old sessions to become invalid
        server_1.stop(None)
        server_1, port_1 = start_secure_server()

        # Old sessions should no longer be valid
        self._do_one_shot_client_rpc(
            channel_creds,
            channel_options,
            port_1,
            expect_ssl_session_reused=[b"false"],
        )

        # Resumption should work for subsequent connections
        self._do_one_shot_client_rpc(
            channel_creds,
            channel_options,
            port_1,
            expect_ssl_session_reused=[b"true"],
        )
        server_1.stop(None)


if __name__ == "__main__":
    logging.basicConfig()
    unittest.main(verbosity=2)



#!/usr/bin/env python3

# Copyright 2016 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import subprocess
import sys

import yaml

errors = 0

os.chdir(os.path.join(os.path.dirname(sys.argv[0]), "../../.."))

# hack import paths to pick up extra code
sys.path.insert(0, os.path.abspath("tools/buildgen/plugins"))
from expand_version import Version

try:
    branch_name = subprocess.check_output(
        "git rev-parse --abbrev-ref HEAD", shell=True
    ).decode()
except:
    print("WARNING: not a git repository")
    branch_name = None

if branch_name is not None:
    m = re.match(r"^release-([0-9]+)_([0-9]+)$", branch_name)
    if m:
        print("RELEASE branch")
        # version number should align with the branched version
        check_version = lambda version: (
            version.major == int(m.group(1))
            and version.minor == int(m.group(2))
        )
        warning = (
            'Version key "%%s" value "%%s" should have a major version %s and'
            " minor version %s" % (m.group(1), m.group(2))
        )
    elif re.match(r"^debian/.*$", branch_name):
        # no additional version checks for debian branches
        check_version = lambda version: True
    else:
        # all other branches should have a -dev tag
        check_version = lambda version: version.tag == "dev"
        warning = 'Version key "%s" value "%s" should have a -dev tag'
else:
    check_version = lambda version: True

with open("build_handwritten.yaml", "r") as f:
    build_yaml = yaml.safe_load(f.read())

settings = build_yaml["settings"]

top_version = Version(settings["version"])
if not check_version(top_version):
    errors += 1
    print((warning % ("version", top_version)))

for tag, value in list(settings.items()):
    if re.match(r"^[a-z]+_version$", tag):
        value = Version(value)
        if tag != "core_version":
            if value.major != top_version.major:
                errors += 1
                print(
                    "major version mismatch on %s: %d vs %d"
                    % (tag, value.major, top_version.major)
                )
            if value.minor != top_version.minor:
                errors += 1
                print(
                    "minor version mismatch on %s: %d vs %d"
                    % (tag, value.minor, top_version.minor)
                )
        if not check_version(value):
            errors += 1
            print((warning % (tag, value)))

sys.exit(errors)



# Copyright 2020 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import dataclasses
import logging
from typing import Any, Dict, List, Optional, Tuple

from google.rpc import code_pb2
import tenacity

from framework.infrastructure import gcp

logger = logging.getLogger(__name__)

# Type aliases
GcpResource = gcp.compute.ComputeV1.GcpResource


@dataclasses.dataclass(frozen=True)
class EndpointPolicy:
    url: str
    name: str
    type: str
    traffic_port_selector: dict
    endpoint_matcher: dict
    update_time: str
    create_time: str
    http_filters: Optional[dict] = None
    server_tls_policy: Optional[str] = None

    @classmethod
    def from_response(
        cls, name: str, response: Dict[str, Any]
    ) -> "EndpointPolicy":
        return cls(
            name=name,
            url=response["name"],
            type=response["type"],
            server_tls_policy=response.get("serverTlsPolicy", None),
            traffic_port_selector=response["trafficPortSelector"],
            endpoint_matcher=response["endpointMatcher"],
            http_filters=response.get("httpFilters", None),
            update_time=response["updateTime"],
            create_time=response["createTime"],
        )


@dataclasses.dataclass(frozen=True)
class Mesh:
    name: str
    url: str
    routes: Optional[List[str]]

    @classmethod
    def from_response(cls, name: str, d: Dict[str, Any]) -> "Mesh":
        return cls(
            name=name,
            url=d["name"],
            routes=list(d["routes"]) if "routes" in d else None,
        )


@dataclasses.dataclass(frozen=True)
class GrpcRoute:
    @dataclasses.dataclass(frozen=True)
    class MethodMatch:
        type: Optional[str]
        grpc_service: Optional[str]
        grpc_method: Optional[str]
        case_sensitive: Optional[bool]

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "GrpcRoute.MethodMatch":
            return cls(
                type=d.get("type"),
                grpc_service=d.get("grpcService"),
                grpc_method=d.get("grpcMethod"),
                case_sensitive=d.get("caseSensitive"),
            )

    @dataclasses.dataclass(frozen=True)
    class HeaderMatch:
        type: Optional[str]
        key: str
        value: str

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "GrpcRoute.HeaderMatch":
            return cls(
                type=d.get("type"),
                key=d["key"],
                value=d["value"],
            )

    @dataclasses.dataclass(frozen=True)
    class RouteMatch:
        method: Optional["GrpcRoute.MethodMatch"]
        headers: Tuple["GrpcRoute.HeaderMatch"]

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "GrpcRoute.RouteMatch":
            return cls(
                method=GrpcRoute.MethodMatch.from_response(d["method"])
                if "method" in d
                else None,
                headers=tuple(
                    GrpcRoute.HeaderMatch.from_response(h) for h in d["headers"]
                )
                if "headers" in d
                else (),
            )

    @dataclasses.dataclass(frozen=True)
    class Destination:
        service_name: str
        weight: Optional[int]

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "GrpcRoute.Destination":
            return cls(
                service_name=d["serviceName"],
                weight=d.get("weight"),
            )

    @dataclasses.dataclass(frozen=True)
    class RouteAction:
        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "GrpcRoute.RouteAction":
            destinations = (
                [
                    GrpcRoute.Destination.from_response(dest)
                    for dest in d["destinations"]
                ]
                if "destinations" in d
                else []
            )
            return cls(destinations=destinations)

    @dataclasses.dataclass(frozen=True)
    class RouteRule:
        matches: List["GrpcRoute.RouteMatch"]
        action: "GrpcRoute.RouteAction"

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "GrpcRoute.RouteRule":
            matches = (
                [GrpcRoute.RouteMatch.from_response(m) for m in d["matches"]]
                if "matches" in d
                else []
            )
            return cls(
                matches=matches,
                action=GrpcRoute.RouteAction.from_response(d["action"]),
            )

    name: str
    url: str
    hostnames: Tuple[str]
    rules: Tuple["GrpcRoute.RouteRule"]
    meshes: Optional[Tuple[str]]

    @classmethod
    def from_response(
        cls, name: str, d: Dict[str, Any]
    ) -> "GrpcRoute.RouteRule":
        return cls(
            name=name,
            url=d["name"],
            hostnames=tuple(d["hostnames"]),
            rules=tuple(d["rules"]),
            meshes=None if d.get("meshes") is None else tuple(d["meshes"]),
        )


@dataclasses.dataclass(frozen=True)
class HttpRoute:
    @dataclasses.dataclass(frozen=True)
    class MethodMatch:
        type: Optional[str]
        http_service: Optional[str]
        http_method: Optional[str]
        case_sensitive: Optional[bool]

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "HttpRoute.MethodMatch":
            return cls(
                type=d.get("type"),
                http_service=d.get("httpService"),
                http_method=d.get("httpMethod"),
                case_sensitive=d.get("caseSensitive"),
            )

    @dataclasses.dataclass(frozen=True)
    class HeaderMatch:
        type: Optional[str]
        key: str
        value: str

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "HttpRoute.HeaderMatch":
            return cls(
                type=d.get("type"),
                key=d["key"],
                value=d["value"],
            )

    @dataclasses.dataclass(frozen=True)
    class RouteMatch:
        method: Optional["HttpRoute.MethodMatch"]
        headers: Tuple["HttpRoute.HeaderMatch"]

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "HttpRoute.RouteMatch":
            return cls(
                method=HttpRoute.MethodMatch.from_response(d["method"])
                if "method" in d
                else None,
                headers=tuple(
                    HttpRoute.HeaderMatch.from_response(h) for h in d["headers"]
                )
                if "headers" in d
                else (),
            )

    @dataclasses.dataclass(frozen=True)
    class Destination:
        service_name: str
        weight: Optional[int]

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "HttpRoute.Destination":
            return cls(
                service_name=d["serviceName"],
                weight=d.get("weight"),
            )

    @dataclasses.dataclass(frozen=True)
    class RouteAction:
        destinations: List["HttpRoute.Destination"]
        stateful_session_affinity: Optional["HttpRoute.StatefulSessionAffinity"]

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "HttpRoute.RouteAction":
            destinations = (
                [
                    HttpRoute.Destination.from_response(dest)
                    for dest in d["destinations"]
                ]
                if "destinations" in d
                else []
            )
            stateful_session_affinity = (
                HttpRoute.StatefulSessionAffinity.from_response(
                    d["statefulSessionAffinity"]
                )
                if "statefulSessionAffinity" in d
                else None
            )
            return cls(
                destinations=destinations,
                stateful_session_affinity=stateful_session_affinity,
            )

    @dataclasses.dataclass(frozen=True)
    class StatefulSessionAffinity:
        cookie_ttl: Optional[str]

        @classmethod
        def from_response(
            cls, d: Dict[str, Any]
        ) -> "HttpRoute.StatefulSessionAffinity":
            return cls(cookie_ttl=d.get("cookieTtl"))

    @dataclasses.dataclass(frozen=True)
    class RouteRule:
        matches: List["HttpRoute.RouteMatch"]
        action: "HttpRoute.RouteAction"

        @classmethod
        def from_response(cls, d: Dict[str, Any]) -> "HttpRoute.RouteRule":
            matches = (
                [HttpRoute.RouteMatch.from_response(m) for m in d["matches"]]
                if "matches" in d
                else []
            )
            return cls(
                matches=matches,
                action=HttpRoute.RouteAction.from_response(d["action"]),
            )

    name: str
    url: str
    hostnames: Tuple[str]
    rules: Tuple["HttpRoute.RouteRule"]
    meshes: Optional[Tuple[str]]

    @classmethod
    def from_response(cls, name: str, d: Dict[str, Any]) -> "HttpRoute":
        return cls(
            name=name,
            url=d["name"],
            hostnames=tuple(d["hostnames"]),
            rules=tuple(d["rules"]),
            meshes=None if d.get("meshes") is None else tuple(d["meshes"]),
        )


class _NetworkServicesBase(
    gcp.api.GcpStandardCloudApiResource, metaclass=abc.ABCMeta
):
    """Base class for NetworkServices APIs."""

    # TODO(https://github.com/grpc/grpc/issues/29532) remove pylint disable
    # pylint: disable=abstract-method

    def __init__(self, api_manager: gcp.api.GcpApiManager, project: str):
        super().__init__(api_manager.networkservices(self.api_version), project)
        # Shortcut to projects/*/locations/ endpoints
        self._api_locations = self.api.projects().locations()

    @property
    def api_name(self) -> str:
        return "networkservices"

    def _execute(
        self, *args, **kwargs
    ):  # pylint: disable=signature-differs,arguments-differ
        # Workaround TD bug: throttled operations are reported as internal.
        # Ref b/175345578
        retryer = tenacity.Retrying(
            retry=tenacity.retry_if_exception(self._operation_internal_error),
            wait=tenacity.wait_fixed(10),
            stop=tenacity.stop_after_delay(5 * 60),
            before_sleep=tenacity.before_sleep_log(logger, logging.DEBUG),
            reraise=True,
        )
        retryer(super()._execute, *args, **kwargs)

    @staticmethod
    def _operation_internal_error(exception):
        return (
            isinstance(exception, gcp.api.OperationError)
            and exception.error.code == code_pb2.INTERNAL
        )


class NetworkServicesV1Beta1(_NetworkServicesBase):
    """NetworkServices API v1beta1."""

    ENDPOINT_POLICIES = "endpointPolicies"

    @property
    def api_version(self) -> str:
        return "v1beta1"

    def create_endpoint_policy(self, name, body: dict) -> GcpResource:
        return self._create_resource(
            collection=self._api_locations.endpointPolicies(),
            body=body,
            endpointPolicyId=name,
        )

    def get_endpoint_policy(self, name: str) -> EndpointPolicy:
        response = self._get_resource(
            collection=self._api_locations.endpointPolicies(),
            full_name=self.resource_full_name(name, self.ENDPOINT_POLICIES),
        )
        return EndpointPolicy.from_response(name, response)

    def delete_endpoint_policy(self, name: str) -> bool:
        return self._delete_resource(
            collection=self._api_locations.endpointPolicies(),
            full_name=self.resource_full_name(name, self.ENDPOINT_POLICIES),
        )


class NetworkServicesV1Alpha1(NetworkServicesV1Beta1):
    """NetworkServices API v1alpha1.

    Note: extending v1beta1 class presumes that v1beta1 is just a v1alpha1 API
    graduated into a more stable version. This is true in most cases. However,
    v1alpha1 class can always override and reimplement incompatible methods.
    """

    HTTP_ROUTES = "httpRoutes"
    GRPC_ROUTES = "grpcRoutes"
    MESHES = "meshes"

    @property
    def api_version(self) -> str:
        return "v1alpha1"

    def create_mesh(self, name: str, body: dict) -> GcpResource:
        return self._create_resource(
            collection=self._api_locations.meshes(), body=body, meshId=name
        )

    def get_mesh(self, name: str) -> Mesh:
        full_name = self.resource_full_name(name, self.MESHES)
        result = self._get_resource(
            collection=self._api_locations.meshes(), full_name=full_name
        )
        return Mesh.from_response(name, result)

    def delete_mesh(self, name: str) -> bool:
        return self._delete_resource(
            collection=self._api_locations.meshes(),
            full_name=self.resource_full_name(name, self.MESHES),
        )

    def create_grpc_route(self, name: str, body: dict) -> GcpResource:
        return self._create_resource(
            collection=self._api_locations.grpcRoutes(),
            body=body,
            grpcRouteId=name,
        )

    def create_http_route(self, name: str, body: dict) -> GcpResource:
        return self._create_resource(
            collection=self._api_locations.httpRoutes(),
            body=body,
            httpRouteId=name,
        )

    def get_grpc_route(self, name: str) -> GrpcRoute:
        full_name = self.resource_full_name(name, self.GRPC_ROUTES)
        result = self._get_resource(
            collection=self._api_locations.grpcRoutes(), full_name=full_name
        )
        return GrpcRoute.from_response(name, result)

    def get_http_route(self, name: str) -> GrpcRoute:
        full_name = self.resource_full_name(name, self.HTTP_ROUTES)
        result = self._get_resource(
            collection=self._api_locations.httpRoutes(), full_name=full_name
        )
        return HttpRoute.from_response(name, result)

    def delete_grpc_route(self, name: str) -> bool:
        return self._delete_resource(
            collection=self._api_locations.grpcRoutes(),
            full_name=self.resource_full_name(name, self.GRPC_ROUTES),
        )

    def delete_http_route(self, name: str) -> bool:
        return self._delete_resource(
            collection=self._api_locations.httpRoutes(),
            full_name=self.resource_full_name(name, self.HTTP_ROUTES),
        )



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Constants and functions for data used in testing."""

import os
import pkgutil

_ROOT_CERTIFICATES_RESOURCE_PATH = "credentials/ca.pem"
_PRIVATE_KEY_RESOURCE_PATH = "credentials/server1.key"
_CERTIFICATE_CHAIN_RESOURCE_PATH = "credentials/server1.pem"


def test_root_certificates():
    return pkgutil.get_data(__name__, _ROOT_CERTIFICATES_RESOURCE_PATH)


def private_key():
    return pkgutil.get_data(__name__, _PRIVATE_KEY_RESOURCE_PATH)


def certificate_chain():
    return pkgutil.get_data(__name__, _CERTIFICATE_CHAIN_RESOURCE_PATH)


def cert_hier_1_root_ca_cert():
    return pkgutil.get_data(
        __name__, "credentials/certificate_hierarchy_1/certs/ca.cert.pem"
    )


def cert_hier_1_intermediate_ca_cert():
    return pkgutil.get_data(
        __name__,
        "credentials/certificate_hierarchy_1/intermediate/certs/intermediate.cert.pem",
    )


def cert_hier_1_client_1_key():
    return pkgutil.get_data(
        __name__,
        "credentials/certificate_hierarchy_1/intermediate/private/client.key.pem",
    )


def cert_hier_1_client_1_cert():
    return pkgutil.get_data(
        __name__,
        "credentials/certificate_hierarchy_1/intermediate/certs/client.cert.pem",
    )


def cert_hier_1_server_1_key():
    return pkgutil.get_data(
        __name__,
        "credentials/certificate_hierarchy_1/intermediate/private/localhost-1.key.pem",
    )


def cert_hier_1_server_1_cert():
    return pkgutil.get_data(
        __name__,
        "credentials/certificate_hierarchy_1/intermediate/certs/localhost-1.cert.pem",
    )


def cert_hier_2_root_ca_cert():
    return pkgutil.get_data(
        __name__, "credentials/certificate_hierarchy_2/certs/ca.cert.pem"
    )


def cert_hier_2_intermediate_ca_cert():
    return pkgutil.get_data(
        __name__,
        "credentials/certificate_hierarchy_2/intermediate/certs/intermediate.cert.pem",
    )


def cert_hier_2_client_1_key():
    return pkgutil.get_data(
        __name__,
        "credentials/certificate_hierarchy_2/intermediate/private/client.key.pem",
    )


def cert_hier_2_client_1_cert():
    return pkgutil.get_data(
        __name__,
        "credentials/certificate_hierarchy_2/intermediate/certs/client.cert.pem",
    )


def cert_hier_2_server_1_key():
    return pkgutil.get_data(
        __name__,
        "credentials/certificate_hierarchy_2/intermediate/private/localhost-1.key.pem",
    )


def cert_hier_2_server_1_cert():
    return pkgutil.get_data(
        __name__,
        "credentials/certificate_hierarchy_2/intermediate/certs/localhost-1.cert.pem",
    )




import cython

from .._cache cimport DNSCache
from .._dns cimport DNSAddress, DNSPointer, DNSQuestion, DNSRecord, DNSRRSet
from .._history cimport QuestionHistory
from .._protocol.incoming cimport DNSIncoming
from .._services.info cimport ServiceInfo
from .._services.registry cimport ServiceRegistry
from .answers cimport QuestionAnswers


cdef object TYPE_CHECKING
cdef cython.uint _ONE_SECOND, _TYPE_PTR, _TYPE_ANY, _TYPE_A, _TYPE_AAAA, _TYPE_SRV, _TYPE_TXT
cdef str _SERVICE_TYPE_ENUMERATION_NAME
cdef cython.set _RESPOND_IMMEDIATE_TYPES
cdef cython.set _ADDRESS_RECORD_TYPES
cdef object IPVersion
cdef object _TYPE_PTR, _CLASS_IN, _DNS_OTHER_TTL

cdef class _QueryResponse:

    cdef bint _is_probe
    cdef cython.list _questions
    cdef float _now
    cdef DNSCache _cache
    cdef cython.dict _additionals
    cdef cython.set _ucast
    cdef cython.set _mcast_now
    cdef cython.set _mcast_aggregate
    cdef cython.set _mcast_aggregate_last_second

    @cython.locals(record=DNSRecord)
    cdef add_qu_question_response(self, cython.dict answers)

    cdef add_ucast_question_response(self, cython.dict answers)

    @cython.locals(answer=DNSRecord, question=DNSQuestion)
    cdef add_mcast_question_response(self, cython.dict answers)

    @cython.locals(maybe_entry=DNSRecord)
    cdef bint _has_mcast_within_one_quarter_ttl(self, DNSRecord record)

    @cython.locals(maybe_entry=DNSRecord)
    cdef bint _has_mcast_record_in_last_second(self, DNSRecord record)

    cdef QuestionAnswers answers(self)

cdef class QueryHandler:

    cdef ServiceRegistry registry
    cdef DNSCache cache
    cdef QuestionHistory question_history

    @cython.locals(service=ServiceInfo)
    cdef _add_service_type_enumeration_query_answers(self, cython.dict answer_set, DNSRRSet known_answers)

    @cython.locals(service=ServiceInfo)
    cdef _add_pointer_answers(self, str lower_name, cython.dict answer_set, DNSRRSet known_answers)

    @cython.locals(service=ServiceInfo, dns_address=DNSAddress)
    cdef _add_address_answers(self, str lower_name, cython.dict answer_set, DNSRRSet known_answers, object type_)

    @cython.locals(question_lower_name=str, type_=cython.uint, service=ServiceInfo)
    cdef cython.dict _answer_question(self, DNSQuestion question, DNSRRSet known_answers)

    @cython.locals(
        msg=DNSIncoming,
        question=DNSQuestion,
        answer_set=cython.dict,
        known_answers=DNSRRSet,
        known_answers_set=cython.set,
        is_probe=object,
        now=object
    )
    cpdef async_response(self, cython.list msgs, cython.bint unicast_source)



#!/usr/bin/env python3
# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Simple Mako renderer.

Just a wrapper around the mako rendering library.
"""

import getopt
import glob
import importlib.util
import os
import pickle
import shutil
import sys
from typing import List

from mako import exceptions
from mako.lookup import TemplateLookup
from mako.runtime import Context
from mako.template import Template
import yaml

PROJECT_ROOT = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "..", ".."
)
# TODO(lidiz) find a better way for plugins to reference each other
sys.path.append(os.path.join(PROJECT_ROOT, "tools", "buildgen", "plugins"))


def out(msg: str) -> None:
    print(msg, file=sys.stderr)


def showhelp() -> None:
    out(
        "mako-renderer.py [-o out] [-m cache] [-P preprocessed_input] [-d dict]"
        " [-d dict...] [-t template] [-w preprocessed_output]"
    )


def render_template(template: Template, context: Context) -> None:
    """Render the mako template with given context.

    Prints an error template to indicate where and what in the template caused
    the render failure.
    """
    try:
        template.render_context(context)
    except:
        out(exceptions.text_error_template().render())
        raise


def main(argv: List[str]) -> None:
    got_input = False
    module_directory = None
    preprocessed_output = None
    dictionary = {}
    json_dict = {}
    got_output = False
    output_name = None
    got_preprocessed_input = False
    output_merged = None

    try:
        opts, args = getopt.getopt(argv, "hM:m:o:t:P:")
    except getopt.GetoptError:
        out("Unknown option")
        showhelp()
        sys.exit(2)

    for opt, arg in opts:
        if opt == "-h":
            out("Displaying showhelp")
            showhelp()
            sys.exit()
        elif opt == "-o":
            if got_output:
                out("Got more than one output")
                showhelp()
                sys.exit(3)
            got_output = True
            output_name = arg
        elif opt == "-m":
            if module_directory is not None:
                out("Got more than one cache directory")
                showhelp()
                sys.exit(4)
            module_directory = arg
        elif opt == "-M":
            if output_merged is not None:
                out("Got more than one output merged path")
                showhelp()
                sys.exit(5)
            output_merged = arg
        elif opt == "-P":
            assert not got_preprocessed_input
            assert json_dict == {}
            with open(arg, "rb") as dict_file:
                dictionary = pickle.load(dict_file)
            got_preprocessed_input = True

    cleared_dir = False
    for arg in args:
        got_input = True
        with open(arg) as f:
            srcs = list(yaml.safe_load_all(f.read()))
        for src in srcs:
            if isinstance(src, str):
                assert len(srcs) == 1
                template = Template(
                    src,
                    filename=arg,
                    module_directory=module_directory,
                    lookup=TemplateLookup(directories=["."]),
                )
                with open(output_name, "w") as output_file:
                    render_template(
                        template, Context(output_file, **dictionary)
                    )
            else:
                # we have optional control data: this template represents
                # a directory
                if not cleared_dir:
                    if not os.path.exists(output_name):
                        pass
                    elif os.path.isfile(output_name):
                        os.unlink(output_name)
                    else:
                        shutil.rmtree(output_name, ignore_errors=True)
                    cleared_dir = True
                items = []
                if "foreach" in src:
                    for el in dictionary[src["foreach"]]:
                        if "cond" in src:
                            args = dict(dictionary)
                            args["selected"] = el
                            if not eval(src["cond"], {}, args):
                                continue
                        items.append(el)
                    assert items
                else:
                    items = [None]
                for item in items:
                    args = dict(dictionary)
                    args["selected"] = item
                    item_output_name = os.path.join(
                        output_name, Template(src["output_name"]).render(**args)
                    )
                    if not os.path.exists(os.path.dirname(item_output_name)):
                        os.makedirs(os.path.dirname(item_output_name))
                    template = Template(
                        src["template"],
                        filename=arg,
                        module_directory=module_directory,
                        lookup=TemplateLookup(directories=["."]),
                    )
                    with open(item_output_name, "w") as output_file:
                        render_template(template, Context(output_file, **args))

    if not got_input and not preprocessed_output:
        out("Got nothing to do")
        showhelp()


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright 2021 The gRPC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from typing import List

from absl import flags
from absl import logging
from absl.testing import absltest
from google.protobuf import json_format

from framework import xds_k8s_testcase
from framework import xds_url_map_testcase
from framework.helpers import skips

flags.adopt_module_key_flags(xds_k8s_testcase)

# Type aliases
_XdsTestServer = xds_k8s_testcase.XdsTestServer
_XdsTestClient = xds_k8s_testcase.XdsTestClient

_SUBSET_SIZE = 4
_NUM_BACKENDS = 8
_NUM_CLIENTS = 3


class SubsettingTest(xds_k8s_testcase.RegularXdsKubernetesTestCase):
    @staticmethod
    def is_supported(config: skips.TestConfig) -> bool:
        # Subsetting is an experimental feature where most work is done on the
        # server-side. We limit it to only run on master branch to save
        # resources.
        return config.version_gte("master")

    def test_subsetting_basic(self) -> None:
        with self.subTest("00_create_health_check"):
            self.td.create_health_check()

        with self.subTest("01_create_backend_services"):
            self.td.create_backend_service(subset_size=_SUBSET_SIZE)

        with self.subTest("02_create_url_map"):
            self.td.create_url_map(self.server_xds_host, self.server_xds_port)

        with self.subTest("03_create_target_proxy"):
            self.td.create_target_proxy()

        with self.subTest("04_create_forwarding_rule"):
            self.td.create_forwarding_rule(self.server_xds_port)

        test_servers: List[_XdsTestServer]
        with self.subTest("05_start_test_servers"):
            test_servers = self.startTestServers(replica_count=_NUM_BACKENDS)

        with self.subTest("06_add_server_backends_to_backend_services"):
            self.setupServerBackends()

        rpc_distribution = collections.defaultdict(int)
        with self.subTest("07_start_test_client"):
            for i in range(_NUM_CLIENTS):
                # Clean created client pods if there is any.
                if self.client_runner.time_start_requested:
                    # TODO(sergiitk): Speed up by reusing the namespace.
                    self.client_runner.cleanup()

                # Create a test client
                test_client: _XdsTestClient = self.startTestClient(
                    test_servers[0]
                )
                # Validate the number of received endpoints
                config = test_client.csds.fetch_client_status(
                    log_level=logging.INFO
                )
                self.assertIsNotNone(config)
                json_config = json_format.MessageToDict(config)
                parsed = xds_url_map_testcase.DumpedXdsConfig(json_config)
                logging.info(
                    "Client %d received endpoints (len=%s): %s",
                    i,
                    len(parsed.endpoints),
                    parsed.endpoints,
                )
                self.assertLen(parsed.endpoints, _SUBSET_SIZE)
                # Record RPC stats
                lb_stats = self.getClientRpcStats(
                    test_client, _NUM_BACKENDS * 25
                )
                for key, value in lb_stats.rpcs_by_peer.items():
                    rpc_distribution[key] += value

        with self.subTest("08_log_rpc_distribution"):
            server_entries = sorted(
                rpc_distribution.items(), key=lambda x: -x[1]
            )
            # Validate if clients are receiving different sets of backends (3
            # client received a total of 4 unique backends == FAIL, a total of 5
            # unique backends == PASS)
            self.assertGreater(len(server_entries), _SUBSET_SIZE)
            logging.info(
                "RPC distribution (len=%s): %s",
                len(server_entries),
                server_entries,
            )
            peak = server_entries[0][1]
            mean = sum(map(lambda x: x[1], server_entries)) / len(
                server_entries
            )
            logging.info(
                "Peak=%d Mean=%.1f Peak-to-Mean-Ratio=%.2f",
                peak,
                mean,
                peak / mean,
            )


if __name__ == "__main__":
    absltest.main(failfast=True)



#!/usr/bin/env python3
# Copyright 2020 The gRPC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for build file generation scripts."""

import importlib.util
import os
import sys
import types
from typing import Any, List, Mapping, Union


def import_python_module(path: str) -> types.ModuleType:
    """Imports the Python file at the given path, returns a module object."""
    module_name = os.path.basename(path).replace(".py", "")
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


class Bunch(dict):
    """Allows dot-accessible dictionaries."""

    def __init__(self, d: Mapping):
        dict.__init__(self, d)
        self.__dict__.update(d)


def to_bunch(var: Any) -> Any:
    """Converts any kind of variable to a Bunch."""
    if isinstance(var, list):
        return [to_bunch(i) for i in var]
    if isinstance(var, dict):
        ret = {}
        for k, v in list(var.items()):
            if isinstance(v, (list, dict)):
                v = to_bunch(v)
            ret[k] = v
        return Bunch(ret)
    else:
        return var


def merge_json(dst: Union[Mapping, List], add: Union[Mapping, List]) -> None:
    """Merges JSON objects recursively."""
    if isinstance(dst, dict) and isinstance(add, dict):
        for k, v in list(add.items()):
            if k in dst:
                if k.startswith("#"):
                    continue
                merge_json(dst[k], v)
            else:
                dst[k] = v
    elif isinstance(dst, list) and isinstance(add, list):
        dst.extend(add)
    else:
        raise TypeError(
            "Tried to merge incompatible objects %s %s\n\n%r\n\n%r"
            % (type(dst).__name__, type(add).__name__, dst, add)
        )
